@@ -4,6 +4,7 @@ from unittest import mock
 import pytest
 
 import click
+from click._utils import UNSET
 
 
 def test_nargs_star(runner):
@@ -19,15 +20,6 @@ def test_nargs_star(runner):
     assert result.output.splitlines() == ["src=foo.txt|bar.txt", "dst=dir"]
 
 
-def test_argument_unbounded_nargs_cant_have_default(runner):
-    with pytest.raises(TypeError, match="nargs=-1"):
-
-        @click.command()
-        @click.argument("src", nargs=-1, default=["42"])
-        def copy(src):
-            pass
-
-
 def test_nargs_tup(runner):
     @click.command()
     @click.argument("name", nargs=1)
@@ -63,6 +55,15 @@ def test_nargs_tup_composite(runner, opts):
     assert result.output.splitlines() == ["name=peter id=1"]
 
 
+def test_nargs_mismatch_with_tuple_type():
+    with pytest.raises(ValueError, match="nargs.*must be 2.*but it was 3"):
+
+        @click.command()
+        @click.argument("test", type=(str, int), nargs=3)
+        def cli(_):
+            pass
+
+
 def test_nargs_err(runner):
     @click.command()
     @click.argument("x")
@@ -165,6 +166,8 @@ def test_stdout_default(runner):
     result = runner.invoke(inout, [])
     assert not result.exception
     assert result.output == "Foo bar baz\n"
+    assert result.stdout == "Foo bar baz\n"
+    assert not result.stderr
 
 
 @pytest.mark.parametrize(
@@ -198,19 +201,6 @@ def test_nargs_envvar(runner, nargs, value, expect):
         assert result.return_value == expect
 
 
-def test_envvar_flag_value(runner):
-    @click.command()
-    # is_flag is implicitly true
-    @click.option("--upper", flag_value="upper", envvar="UPPER")
-    def cmd(upper):
-        click.echo(upper)
-        return upper
-
-    # For whatever value of the `env` variable, if it exists, the flag should be `upper`
-    result = runner.invoke(cmd, env={"UPPER": "whatever"})
-    assert result.output.strip() == "upper"
-
-
 def test_nargs_envvar_only_if_values_empty(runner):
     @click.command()
     @click.argument("arg", envvar="X", nargs=-1)
@@ -255,13 +245,43 @@ def test_missing_arg(runner):
     assert "Missing argument 'ARG'." in result.output
 
 
-def test_missing_argument_string_cast():
+@pytest.mark.parametrize(
+    ("value", "expect_missing", "processed_value"),
+    [
+        # Unspecified type of the argument fallback to string, so everything is
+        # processed the click.STRING type.
+        ("", False, ""),
+        ("  ", False, "  "),
+        ("foo", False, "foo"),
+        ("12", False, "12"),
+        (12, False, "12"),
+        (12.1, False, "12.1"),
+        (list(), False, "[]"),
+        (tuple(), False, "()"),
+        (set(), False, "set()"),
+        (frozenset(), False, "frozenset()"),
+        (dict(), False, "{}"),
+        # None is a value that is allowed to be processed by a required argument
+        # because at this stage, the process_value method happens after the default is
+        # applied.
+        (None, False, None),
+        # An UNSET required argument will raise MissingParameter.
+        (UNSET, True, None),
+    ],
+)
+def test_required_argument(value, expect_missing, processed_value):
+    """Test how a required argument is processing the provided values."""
     ctx = click.Context(click.Command(""))
+    argument = click.Argument(["a"], required=True)
 
-    with pytest.raises(click.MissingParameter) as excinfo:
-        click.Argument(["a"], required=True).process_value(ctx, None)
+    if expect_missing:
+        with pytest.raises(click.MissingParameter) as excinfo:
+            argument.process_value(ctx, value)
+        assert str(excinfo.value) == "Missing parameter: a"
 
-    assert str(excinfo.value) == "Missing parameter: a"
+    else:
+        value = argument.process_value(ctx, value)
+        assert value == processed_value
 
 
 def test_implicit_non_required(runner):
@@ -355,22 +375,115 @@ def test_nargs_specified_plus_star_ordering(runner):
     assert result.output.splitlines() == ["('a', 'b', 'c')", "d", "('e', 'f')"]
 
 
-def test_defaults_for_nargs(runner):
+@pytest.mark.parametrize(
+    ("argument_params", "args", "expected"),
+    [
+        # Any iterable with the same number of arguments as nargs is valid.
+        [{"nargs": 2, "default": (1, 2)}, [], (1, 2)],
+        [{"nargs": 2, "default": (1.1, 2.2)}, [], (1, 2)],
+        [{"nargs": 2, "default": ("1", "2")}, [], (1, 2)],
+        [{"nargs": 2, "default": (None, None)}, [], (None, None)],
+        [{"nargs": 2, "default": [1, 2]}, [], (1, 2)],
+        [{"nargs": 2, "default": {1, 2}}, [], (1, 2)],
+        [{"nargs": 2, "default": frozenset([1, 2])}, [], (1, 2)],
+        [{"nargs": 2, "default": {1: "a", 2: "b"}}, [], (1, 2)],
+        # Empty iterable is valid if default is None.
+        [{"nargs": 2, "default": None}, [], None],
+        # Arguments overrides the default.
+        [{"nargs": 2, "default": (1, 2)}, ["3", "4"], (3, 4)],
+        # Unbounded arguments are allowed to have a default.
+        # See: https://github.com/pallets/click/issues/2164
+        [{"nargs": -1, "default": [42]}, [], (42,)],
+        [{"nargs": -1, "default": None}, [], ()],
+        [{"nargs": -1, "default": {1, 2, 3, 4, 5}}, [], (1, 2, 3, 4, 5)],
+    ],
+)
+def test_good_defaults_for_nargs(runner, argument_params, args, expected):
     @click.command()
-    @click.argument("a", nargs=2, type=int, default=(1, 2))
+    @click.argument("a", type=int, **argument_params)
     def cmd(a):
-        x, y = a
-        click.echo(x + y)
+        click.echo(repr(a), nl=False)
+
+    result = runner.invoke(cmd, args)
+    assert result.output == repr(expected)
+
+
+@pytest.mark.parametrize(
+    ("default", "message"),
+    [
+        # Non-iterables defaults.
+        ["Yo", "Error: Invalid value for '[A]...': Value must be an iterable."],
+        ["", "Error: Invalid value for '[A]...': Value must be an iterable."],
+        [True, "Error: Invalid value for '[A]...': Value must be an iterable."],
+        [False, "Error: Invalid value for '[A]...': Value must be an iterable."],
+        [12, "Error: Invalid value for '[A]...': Value must be an iterable."],
+        [7.9, "Error: Invalid value for '[A]...': Value must be an iterable."],
+        # Generator default.
+        [(), "Error: Invalid value for '[A]...': Takes 2 values but 0 were given."],
+        # Unset default.
+        [UNSET, "Error: Missing argument 'A...'."],
+        # Tuples defaults with wrong length.
+        [
+            tuple(),
+            "Error: Invalid value for '[A]...': Takes 2 values but 0 were given.",
+        ],
+        [(1,), "Error: Invalid value for '[A]...': Takes 2 values but 1 was given."],
+        [
+            (1, 2, 3),
+            "Error: Invalid value for '[A]...': Takes 2 values but 3 were given.",
+        ],
+        # Lists defaults with wrong length.
+        [list(), "Error: Invalid value for '[A]...': Takes 2 values but 0 were given."],
+        [[1], "Error: Invalid value for '[A]...': Takes 2 values but 1 was given."],
+        [
+            [1, 2, 3],
+            "Error: Invalid value for '[A]...': Takes 2 values but 3 were given.",
+        ],
+        # Sets defaults with wrong length.
+        [set(), "Error: Invalid value for '[A]...': Takes 2 values but 0 were given."],
+        [
+            set([1]),
+            "Error: Invalid value for '[A]...': Takes 2 values but 1 was given.",
+        ],
+        [
+            set([1, 2, 3]),
+            "Error: Invalid value for '[A]...': Takes 2 values but 3 were given.",
+        ],
+        # Frozensets defaults with wrong length.
+        [
+            frozenset(),
+            "Error: Invalid value for '[A]...': Takes 2 values but 0 were given.",
+        ],
+        [
+            frozenset([1]),
+            "Error: Invalid value for '[A]...': Takes 2 values but 1 was given.",
+        ],
+        [
+            frozenset([1, 2, 3]),
+            "Error: Invalid value for '[A]...': Takes 2 values but 3 were given.",
+        ],
+        # Dictionaries defaults with wrong length.
+        [dict(), "Error: Invalid value for '[A]...': Takes 2 values but 0 were given."],
+        [
+            {1: "a"},
+            "Error: Invalid value for '[A]...': Takes 2 values but 1 was given.",
+        ],
+        [
+            {1: "a", 2: "b", 3: "c"},
+            "Error: Invalid value for '[A]...': Takes 2 values but 3 were given.",
+        ],
+    ],
+)
+def test_bad_defaults_for_nargs(runner, default, message):
+    """Some defaults are not valid when nargs is set."""
+
+    @click.command()
+    @click.argument("a", nargs=2, type=int, default=default)
+    def cmd(a):
+        click.echo(repr(a))
 
     result = runner.invoke(cmd, [])
-    assert result.output.strip() == "3"
-
-    result = runner.invoke(cmd, ["3", "4"])
-    assert result.output.strip() == "7"
-
-    result = runner.invoke(cmd, ["3"])
-    assert result.exception is not None
-    assert "Argument 'a' takes 2 values." in result.output
+    assert message in result.stderr
 
 
 def test_multiple_param_decls_not_allowed(runner):
@@ -387,12 +500,6 @@ def test_multiple_not_allowed():
         click.Argument(["a"], multiple=True)
 
 
-@pytest.mark.parametrize("value", [(), ("a",), ("a", "b", "c")])
-def test_nargs_bad_default(runner, value):
-    with pytest.raises(ValueError, match="nargs=2"):
-        click.Argument(["a"], nargs=2, default=value)
-
-
 def test_subcommand_help(runner):
     @click.group()
     @click.argument("name")
@@ -472,3 +579,40 @@ def test_duplicate_names_warning(runner, args_one, args_two):
 
     with pytest.warns(UserWarning):
         runner.invoke(cli, [])
+
+
+@pytest.mark.parametrize(
+    ("argument_kwargs", "pass_argv"),
+    (
+        # there is a large potential parameter space to explore here
+        # this is just a very small sample of it
+        ({}, ["myvalue"]),
+        ({"nargs": -1}, []),
+        ({"nargs": -1}, ["myvalue"]),
+        ({"default": None}, ["myvalue"]),
+        ({"required": False}, []),
+        ({"required": False}, ["myvalue"]),
+    ),
+)
+def test_argument_custom_class_can_override_type_cast_value_and_never_sees_unset(
+    runner, argument_kwargs, pass_argv
+):
+    """
+    Test that overriding type_cast_value is supported
+
+    In particular, the argument is never passed an UNSET sentinel value.
+    """
+
+    class CustomArgument(click.Argument):
+        def type_cast_value(self, ctx, value):
+            assert value is not UNSET
+            return value
+
+    @click.command()
+    @click.argument("myarg", **argument_kwargs, cls=CustomArgument)
+    def cmd(myarg):
+        click.echo("ok")
+
+    result = runner.invoke(cmd, pass_argv)
+    assert not result.exception
+    assert result.exit_code == 0
